function [V,dft,F] = AV_stpp(x,y,basis_x,basis_y,sm,per,param,effects)
% [V,dft,F] = AV_stpp(x,y,basis_x,basis_y,sm,per,param,effects)
%
% Asymptotic covariance matrix and drift of sqrt(n)*(theta_hat-theta),
%   where theta_hat are the STPP parameter estimators:
%   (Szu,Szv,vec(Suv),tau,s2z,c0,vec(C),s2u,d0,vec(D),s2v)
%
% Note: the estimated covariance matrix of the estimators is 
%       V/n and the bias is dft/sqrt(n)
%
% INPUT:
%   x, y, basis_x, basis_y, sm,per: as in STPP
%   param, effects: output of STPP (parameter and random-effect estimators)
%
% OUTPUT:
%   V: Asymptotic covariance matrix (d x d)
%   drf: Asymptotic drift (d x 1)
%   F: Fisher's information matrix (d x d)
%
% External calls: BSPL, NGBASIS
%
% Version: February 2019

%% Data filtering (elimination of data outside RNG)
n = length(x);
    % Filtering X's (and associated Y's)
a = basis_x.rng(1);
b = basis_x.rng(2);
for i = 1:n
    jout = x{i}<a | x{i}>b;
    x{i}(jout) = [];
    y{i}(jout,:) = [];
end
    % Filtering Y's (and associated X's)
m = zeros(n,1);
for i = 1:n
    jout = ~basis_y.Rdom(y{i});
    y{i}(jout,:) = [];
    x{i}(jout) = [];
    m(i) = length(x{i});
end

%% Initialization
    % X basis quantities
ng1 = 300;
q1 = basis_x.or + basis_x.nk;
t = linspace(a,b,ng1);
dt = t(2)-t(1);
knt = linspace(a,b,basis_x.nk+2);
B0_x = bspl(t,basis_x.or,knt,0);
J0_x = (B0_x'*B0_x)*dt;
if basis_x.or>2
    B2_x = bspl(t,basis_x.or,knt,2);
    J2_x = (B2_x'*B2_x)*dt;
else
    J2_x = zeros(q1,q1);
end
sumB_x = zeros(n,q1);
for i = 1:n
    B_x_i = bspl(x{i},basis_x.or,knt,0);
    sumB_x(i,:) = sum(B_x_i,1);
end
    % Y basis quantities
[J0_y,J2_y] = compJ(basis_y);
ng2 = 100;
s1 = linspace(basis_y.I1(1),basis_y.I1(2),ng2)';
s2 = linspace(basis_y.I2(1),basis_y.I2(2),ng2)';
ds1 = s1(2)-s1(1);
ds2 = s2(2)-s2(1);
ds = ds1*ds2;
s = [kron(s1,ones(ng2,1)) kron(ones(ng2,1),s2)];
s = s(basis_y.Rdom(s),:);
B0_y = NGbasis(s,basis_y,0);
q2 = size(B0_y,2);
sumB_y = zeros(n,q2);
for i = 1:n
    B_y_i = NGbasis(y{i},basis_y,0);
    sumB_y(i,:) = sum(B_y_i,1);
end

%% Fisher's information matrix
S = [param.s2z, param.Szu, param.Szv; ...
    param.Szu', diag(param.s2u), param.Suv; ...
    param.Szv', param.Suv', diag(param.s2v)];
w = [effects.z,effects.u,effects.v];
[w,w2] = compeff(sumB_x,sumB_y,param.tau,param.c0,param.d0,param.C,...
    param.D,S,dt,ds,B0_x,B0_y,m,w);
theta = [param.Szu(:);param.Szv(:);vec(param.Suv);param.tau;param.s2z;...
    param.c0;vec(param.C);param.s2u;param.d0;vec(param.D);param.s2v];
p1 = size(param.C,2);
p2 = size(param.D,2);
d1 = p1;
d2 = d1 + p2;
d3 = d2 + p1*p2;
d4 = d3 + 1;
d5 = d4 + 1;
d6 = d5 + q1;
d7 = d6 + q1*p1;
d8 = d7 + p1;
d9 = d8 + q2;
d10 = d9 + q2*p2;
d_full = d10 + p2;
Dlogf = zeros(d_full,n);
S1 = inv(S);
iu = 2:1+p1;
iv = p1+2:1+p1+p2;
S1_11 = S1(1,1); S1_12 = S1(1,iu); S1_13 = S1(1,iv);
S1_21 = S1(iu,1); S1_22 = S1(iu,iu); S1_23 = S1(iu,iv);
S1_31 = S1(iv,1); S1_32 = S1(iv,iu); S1_33 = S1(iv,iv);

for i = 1:n
    z = w(i,1);
    u = w(i,iu)';
    v = w(i,iv)';
    lmb_x = exp(B0_x*param.c0+B0_x*param.C*u);
    lmb_y = exp(B0_y*param.d0+B0_y*param.D*v);
    r = exp(param.tau+z);
    I_x = sum(lmb_x)*dt;
    I_y = sum(lmb_y)*ds;
    Dlogf(1:d1,i) = -vec(S1_21) + ...
        vec([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_11;S1_21;S1_31]);
    Dlogf(d1+1:d2,i) = -vec(S1_31) + ...
        vec([S1_31,S1_32,S1_33]*w2(:,:,i)*[S1_11;S1_21;S1_31]);
    Dlogf(d2+1:d3,i) = -vec(S1_23) + ...
        vec([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_13;S1_23;S1_33]);
    Dlogf(d3+1:d4,i) = -r*I_x*I_y + m(i);
    Dlogf(d4+1:d5,i) = -0.5*S1_11 + ...
        0.5*[S1_11,S1_12,S1_13]*w2(:,:,i)*[S1_11;S1_21;S1_31];
    Dlogf(d5+1:d6,i) = -r*I_y*(B0_x'*lmb_x)*dt + sumB_x(i,:)';
    Dlogf(d6+1:d7,i) = ...
        -r*I_y*kron(u,(B0_x'*lmb_x)*dt) + kron(u,sumB_x(i,:)');
    Dlogf(d7+1:d8,i) = -0.5*diag(S1_22) + ...
        0.5*diag([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_12;S1_22;S1_32]);
    Dlogf(d8+1:d9,i) = -r*I_x*(B0_y'*lmb_y)*ds + sumB_y(i,:)';
    Dlogf(d9+1:d10,i) = ...
        -r*I_x*kron(v,(B0_y'*lmb_y)*ds) + kron(v,sumB_y(i,:)');
    Dlogf(d10+1:d_full,i) = -0.5*diag(S1_33) + ...
        0.5*diag([S1_31,S1_32,S1_33]*w2(:,:,i)*[S1_13;S1_23;S1_33]);
end
F = (Dlogf*Dlogf')/n;

% Derivatives of constraints
Bab0 = bspl([a b],basis_x.or,knt,0);
Bab1 = bspl([a b],basis_x.or,knt,1);
a_P0 = (Bab0(1,:)-Bab0(2,:))';
a_P1 = (Bab1(1,:)-Bab1(2,:))';
a_t0 = (sum(B0_x,1)*dt)';
a_s0 = (sum(B0_y,1)*ds)';
A_C_orth = zeros(p1*(p1+1)/2,d_full);
A_C_int0 = zeros(p1+1,d_full);
A_C_per0 = zeros(p1+1,d_full);
A_C_per1 = zeros(p1+1,d_full);
A_D_orth = zeros(p2*(p2+1)/2,d_full);
A_D_int0 = zeros(p2+1,d_full);
    % Orth const Cs
row = 0;
for k = 1:p1
    K_k = zeros(q1,d_full);
    K_k(:,d6+(k-1)*q1+1:d6+k*q1) = eye(q1);
    for l = 1:k
        K_l = zeros(q1,d_full);
        K_l(:,d6+(l-1)*q1+1:d6+l*q1) = eye(q1);
        row = row + 1;
        A_C_orth(row,:) = theta'*(K_k'*J0_x*K_l + K_l'*J0_x*K_k);
    end
end
    % Zero-integ and periodicity const Cs
row = 0;
for k = 0:p1
    K_k = zeros(q1,d_full);
    if k==0
        K_k(:,d5+1:d5+q1) = eye(q1);
    else
        K_k(:,d6+(k-1)*q1+1:d6+k*q1) = eye(q1);
    end
    row = row + 1;
    A_C_int0(row,:) = a_t0'*K_k;
    A_C_per0(row,:) = a_P0'*K_k;
    A_C_per1(row,:) = a_P1'*K_k;
end
    % Orth const Ds
row = 0;
for k = 1:p2
    K_k = zeros(q2,d_full);
    K_k(:,d9+(k-1)*q2+1:d9+k*q2) = eye(q2);
    for l = 1:k
        K_l = zeros(q2,d_full);
        K_l(:,d9+(l-1)*q2+1:d9+l*q2) = eye(q2);
        row = row + 1;
        A_D_orth(row,:) = theta'*(K_k'*J0_y*K_l + K_l'*J0_y*K_k);
    end
end
    % Zero-integ const Ds
row = 0;
for k = 0:p2
    K_k = zeros(q2,d_full);
    if k==0
        K_k(:,d8+1:d8+q2) = eye(q2);
    else
        K_k(:,d9+(k-1)*q2+1:d9+k*q2) = eye(q2);
    end
    row = row + 1;
    A_D_int0(row,:) = a_s0'*K_k;
end
A = [A_C_orth;A_C_int0;A_D_orth;A_D_int0];
if per
    A = [A;A_C_per0;A_C_per1];
end

% Derivatives of smoothness penalties
K_c0 = zeros(q1,d_full);
K_c0(:,d5+1:d6) = eye(q1);
K_vecC = zeros(q1*p1,d_full);
K_vecC(:,d6+1:d7) = eye(q1*p1);
K_d0 = zeros(q2,d_full);
K_d0(:,d8+1:d9) = eye(q2);
K_vecD = zeros(q2*p2,d_full);
K_vecD(:,d9+1:d10) = eye(q2*p2);
DP = [2*theta'*K_c0'*J2_x*K_c0; ...
    2*theta'*K_vecC'*kron(eye(p1),J2_x)*K_vecC; ...
    2*theta'*K_d0'*J2_y*K_d0; ...
    2*theta'*K_vecD'*kron(eye(p2),J2_y)*K_vecD];

% Asymptotic variance and drift
kappa = sqrt(n)*[sm.muX; sm.pcX; sm.muY; sm.pcY];
B = null(A)';
V = B'*((B*F*B')\B);
dft = -V*DP'*kappa;
end

%%%%%%% ------------ AUXILIARY FUNCTIONS ------------

function v = vec(A)
v = A(:);
end

function [w,w2,logf,flag] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
    S,dt,ds,B0_x,B0_y,m,w_ini)
%Computes random effects and log-pdf's using Laplace approximations
n = size(sumB_x,1);
p1 = size(C,2);
p2 = size(D,2);
p = p1+p2+1;
Phi_x = B0_x*C;
Phi_y = B0_y*D;
logf = zeros(n,1);
w = w_ini;
w2 = zeros(p,p,n);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    gi = -Inf;
    for steps = 1:NR_steps
        if rcond(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            zL = wL(1);
            uL = wL(2:p1+1);
            vL = wL(p1+2:p);
            r = exp(tau+zL);
            lmbi_x = exp(B0_x*(c0+C*uL'));
            lmbi_y = exp(B0_y*(d0+D*vL'));
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            I_x_der = (lmbi_x'*Phi_x)*dt;
            I_y_der = (lmbi_y'*Phi_y)*ds;
            I_x_hess = Phi_x'*((lmbi_x*ones(1,p1)).*Phi_x)*dt;
            I_y_hess = Phi_y'*((lmbi_y*ones(1,p2)).*Phi_y)*ds;
            D_gi = [-r*I_x*I_y + m(i),...
                -r*I_y*I_x_der + sumB_x(i,:)*C, ...
                -r*I_x*I_y_der + sumB_y(i,:)*D] ...
                - wL/S;
            H_gi = [-r*I_x*I_y, -r*I_y*I_x_der, -r*I_x*I_y_der; ...
                -r*I_y*I_x_der', -r*I_y*I_x_hess, -r*I_x_der'*I_y_der;... 
                -r*I_x*I_y_der', -r*I_y_der'*I_x_der, -r*I_x*I_y_hess]...
                - S\eye(p);
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*(tau+zL) ...
                + sumB_x(i,:)*(c0+C*uL') + sumB_y(i,:)*(d0+D*vL') ...
                -.5*p*log(2*pi) - .5*log(det(S)) - .5*wL*(S\wL');
        else
            flag = true;
        end
    end
    if rcond(H_gi)>eps
        SL = (-H_gi)\eye(p);
    else
        SL = (-last_H)\eye(p);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (p/2)*log(2*pi) + .5*log(det(SL));
    w(i,:) = wL;
    w2(:,:,i) = SL + wL'*wL;
end
end

function [J0,J2] = compJ(basis)
% Computes matrices J0 and J2 for roughness penalty terms
% J0 is the Gram matrix of the basis functions
% Internal Jik are the Gram matrices of the 2nd deriv wrt (ti,tk)
% J2 = J11 + 2*J12 + J22
m = 100;
g1 = linspace(basis.I1(1),basis.I1(2),m)';
g2 = linspace(basis.I2(1),basis.I2(2),m)';
d1 = g1(2)-g1(1);
d2 = g2(2)-g2(1);
xg_0 = [kron(g1,ones(m,1)) kron(ones(m,1),g2)];
xg_0 = xg_0(basis.Rdom(xg_0),:);
B0_0 = NGbasis(xg_0,basis,0);
q = size(B0_0,2);
J0 = zeros(q,q);
J11 = zeros(q,q);
J12 = zeros(q,q);
J22 = zeros(q,q);
for j = 1:m
    xg = [g1(j)*ones(m,1) g2];
    xg = xg(basis.Rdom(xg),:);
    B0 = NGbasis(xg,basis,0);
    J0 = J0 + d1*d2*(B0'*B0);
    B2 = NGbasis(xg,basis,2);
    J11 = J11 + d1*d2*(B2(:,:,1)'*B2(:,:,1));
    J12 = J12 + d1*d2*(B2(:,:,2)'*B2(:,:,2));
    J22 = J22 + d1*d2*(B2(:,:,3)'*B2(:,:,3));
end
J2 = J11 + 2*J12 + J22;
end
