function [V,F] = AV_stpp_simp(x,y,basis_x,basis_y,param,effects)
% [V,F] = AV_stpp_simp(x,y,basis_x,basis_y,param,effects)
%
% Simplified asymptotic covariance matrix of sqrt(n)*(theta_hat-theta),
%   where theta is the reduced vector of parameters:
%
%   (Szu,Szv,vec(Suv),tau,s2z,s2u,s2v)
%
% The functional parameters are assumed fixed (although plug-in STPP 
%   estimators are used) so as to reduce the dimensionality of theta.
%
% Note: The estimated covariance matrix of theta_hat is V/n.
%
% INPUT:
%   x, y, basis_x, basis_y: as in STPP
%   param, effects: output of STPP (parameter and random-effect estimators)
%
% OUTPUT:
%   V: Asymptotic covariance matrix (d x d)
%   F: Fisher's information matrix (d x d)
%
% External calls: BSPL, NGBASIS
%
% Version: February 2019

%% Data filtering (elimination of data outside RNG)
n = length(x);
    % Filtering X's (and associated Y's)
a = basis_x.rng(1);
b = basis_x.rng(2);
for i = 1:n
    jout = x{i}<a | x{i}>b;
    x{i}(jout) = [];
    y{i}(jout,:) = [];
end
    % Filtering Y's (and associated X's)
m = zeros(n,1);
for i = 1:n
    jout = ~basis_y.Rdom(y{i});
    y{i}(jout,:) = [];
    x{i}(jout) = [];
    m(i) = length(x{i});
end

%% Initialization
    % X basis quantities
ng1 = 300;
q1 = basis_x.or + basis_x.nk;
t = linspace(a,b,ng1);
dt = t(2)-t(1);
knt = linspace(a,b,basis_x.nk+2);
B0_x = bspl(t,basis_x.or,knt,0);
sumB_x = zeros(n,q1);
for i = 1:n
    B_x_i = bspl(x{i},basis_x.or,knt,0);
    sumB_x(i,:) = sum(B_x_i,1);
end
    % Y basis quantities
ng2 = 100;
s1 = linspace(basis_y.I1(1),basis_y.I1(2),ng2)';
s2 = linspace(basis_y.I2(1),basis_y.I2(2),ng2)';
ds1 = s1(2)-s1(1);
ds2 = s2(2)-s2(1);
ds = ds1*ds2;
s = [kron(s1,ones(ng2,1)) kron(ones(ng2,1),s2)];
s = s(basis_y.Rdom(s),:);
B0_y = NGbasis(s,basis_y,0);
q2 = size(B0_y,2);
sumB_y = zeros(n,q2);
for i = 1:n
    B_y_i = NGbasis(y{i},basis_y,0);
    sumB_y(i,:) = sum(B_y_i,1);
end

%% Fisher's information matrix
S = [param.s2z, param.Szu, param.Szv; ...
    param.Szu', diag(param.s2u), param.Suv; ...
    param.Szv', param.Suv', diag(param.s2v)];
w = [effects.z,effects.u,effects.v];
[w,w2] = compeff(sumB_x,sumB_y,param.tau,param.c0,param.d0,param.C,...
    param.D,S,dt,ds,B0_x,B0_y,m,w);
p1 = size(param.C,2);
p2 = size(param.D,2);
d1 = p1;
d2 = d1 + p2;
d3 = d2 + p1*p2;
d4 = d3 + 1;
d5 = d4 + 1;
d6 = d5 + p1;
d_full = d6 + p2;
Dlogf = zeros(d_full,n);
S1 = inv(S);
iu = 2:1+p1;
iv = p1+2:1+p1+p2;
S1_11 = S1(1,1); S1_12 = S1(1,iu); S1_13 = S1(1,iv);
S1_21 = S1(iu,1); S1_22 = S1(iu,iu); S1_23 = S1(iu,iv);
S1_31 = S1(iv,1); S1_32 = S1(iv,iu); S1_33 = S1(iv,iv);

for i = 1:n
    z = w(i,1);
    u = w(i,iu)';
    v = w(i,iv)';
    lmb_x = exp(B0_x*param.c0+B0_x*param.C*u);
    lmb_y = exp(B0_y*param.d0+B0_y*param.D*v);
    r = exp(param.tau+z);
    I_x = sum(lmb_x)*dt;
    I_y = sum(lmb_y)*ds;
    Dlogf(1:d1,i) = -vec(S1_21) + ...
        vec([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_11;S1_21;S1_31]);
    Dlogf(d1+1:d2,i) = -vec(S1_31) + ...
        vec([S1_31,S1_32,S1_33]*w2(:,:,i)*[S1_11;S1_21;S1_31]);
    Dlogf(d2+1:d3,i) = -vec(S1_23) + ...
        vec([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_13;S1_23;S1_33]);
    Dlogf(d3+1:d4,i) = -r*I_x*I_y + m(i);
    Dlogf(d4+1:d5,i) = -0.5*S1_11 + ...
        0.5*[S1_11,S1_12,S1_13]*w2(:,:,i)*[S1_11;S1_21;S1_31];
    Dlogf(d5+1:d6,i) = -0.5*diag(S1_22) + ...
        0.5*diag([S1_21,S1_22,S1_23]*w2(:,:,i)*[S1_12;S1_22;S1_32]);
    Dlogf(d6+1:d_full,i) = -0.5*diag(S1_33) + ...
        0.5*diag([S1_31,S1_32,S1_33]*w2(:,:,i)*[S1_13;S1_23;S1_33]);
end
F = (Dlogf*Dlogf')/n;

%% Asymptotic covariance matrix
V = inv(F);
end

%%%%%%% ------------ AUXILIARY FUNCTIONS ------------

function v = vec(A)
v = A(:);
end

function [w,w2,logf,flag] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
    S,dt,ds,B0_x,B0_y,m,w_ini)
%Computes random effects and log-pdf's using Laplace approximations
n = size(sumB_x,1);
p1 = size(C,2);
p2 = size(D,2);
p = p1+p2+1;
Phi_x = B0_x*C;
Phi_y = B0_y*D;
logf = zeros(n,1);
w = w_ini;
w2 = zeros(p,p,n);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    gi = -Inf;
    for steps = 1:NR_steps
        if rcond(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            zL = wL(1);
            uL = wL(2:p1+1);
            vL = wL(p1+2:p);
            r = exp(tau+zL);
            lmbi_x = exp(B0_x*(c0+C*uL'));
            lmbi_y = exp(B0_y*(d0+D*vL'));
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            I_x_der = (lmbi_x'*Phi_x)*dt;
            I_y_der = (lmbi_y'*Phi_y)*ds;
            I_x_hess = Phi_x'*((lmbi_x*ones(1,p1)).*Phi_x)*dt;
            I_y_hess = Phi_y'*((lmbi_y*ones(1,p2)).*Phi_y)*ds;
            D_gi = [-r*I_x*I_y + m(i),...
                -r*I_y*I_x_der + sumB_x(i,:)*C, ...
                -r*I_x*I_y_der + sumB_y(i,:)*D] ...
                - wL/S;
            H_gi = [-r*I_x*I_y, -r*I_y*I_x_der, -r*I_x*I_y_der; ...
                -r*I_y*I_x_der', -r*I_y*I_x_hess, -r*I_x_der'*I_y_der;... 
                -r*I_x*I_y_der', -r*I_y_der'*I_x_der, -r*I_x*I_y_hess]...
                - S\eye(p);
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*(tau+zL) ...
                + sumB_x(i,:)*(c0+C*uL') + sumB_y(i,:)*(d0+D*vL') ...
                -.5*p*log(2*pi) - .5*log(det(S)) - .5*wL*(S\wL');
        else
            flag = true;
        end
    end
    if rcond(H_gi)>eps
        SL = (-H_gi)\eye(p);
    else
        SL = (-last_H)\eye(p);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (p/2)*log(2*pi) + .5*log(det(SL));
    w(i,:) = wL;
    w2(:,:,i) = SL + wL'*wL;
end
end
