function B = NGbasis(x,param,der)
% B = NGbasis(x,param,der)
%
% Normalized Gaussian radial basis functions (and derivatives)
%
% INPUT: 
%   x: Input points (m x 2 vector).
%   param: Basis parameters. Struct with the following fields:
%       I1: Range for first coordinate of x's (2 x 1 vector).
%       I2: Range for second coordinate of x's (2 x 1 vector).
%           (Note: I1 x I2 must contain the actual domain R,
%            which is defined by Rdom below).
%       Rdom: Domain indicator function (function handle).
%             For an (m x 2) matrix t, rng(t) must return a logical 
%             (m x 1) vector indicating whether each row of t is in the
%             region of interest R.
%       nk: Number of knots on each axis (integer).
%           Equally spaced knots on each axis are used, giving in principle
%           nk^2 centroids on I1 x I2 for the Gaussian kernels, but those
%           outside R are eliminated. As "sigma" for each Gaussian kernel 
%           we take half the distance between its centroid and the next 
%           closest centroid.
%   der: Order of derivative to compute (0, 1 or 2).
% OUTPUT:
%   B: Basis functions (or derivatives) at x. Specifically:
%       If der=0, B is (m x p) with function values.
%       If der=1, B is (m x p x 2), where B(:,:,j) is der wrt xj. 
%       If der=2, B is (m x p x 3), where B(:,:,1) is 2nd der wrt (x1,x1),
%        B(:,:,2) wrt (x1,x2), and B(:,:,3) wrt (x2,x2).
%
% Version: April 2017
%

% Input check
XinR = param.Rdom(x);
if ~all(XinR)
    disp('Warning: some values of X are outside domain R')
    disp('Function will work, but division by 0 may be encountered')
end
    
% Knot definition
kn1 = linspace(param.I1(1),param.I1(2),param.nk)';
kn2 = linspace(param.I2(1),param.I2(2),param.nk)';
knts = [kron(kn1,ones(param.nk,1)) kron(ones(param.nk,1),kn2)];
inR = param.Rdom(knts);
knts(~inR,:) = [];

% Main computations
m = size(x,1);
p = size(knts,1);
nrms2 = zeros(m,p);
sigma2 = zeros(p,1);
for j = 1:p
    nrms2(:,j) = sum((x-ones(m,1)*knts(j,:)).^2,2);
    dist2 = sum((knts([1:j-1,j+1:p],:)-ones(p-1,1)*knts(j,:)).^2,2);
    sigma2(j) = min(dist2)/4;
end
  % Preliminary basis (Gaussian kernels)
B0 = exp(-.5*nrms2*diag(1./sigma2));
if der==1 || der==2     % 1st deriv wrt x1 and x2
    B1(:,:,1) = exp(-.5*nrms2*diag(1./sigma2)).* ...
            (-(x(:,1)*ones(1,p)-ones(m,1)*knts(:,1)')*diag(1./sigma2));
    B1(:,:,2) = exp(-.5*nrms2*diag(1./sigma2)).* ...
            (-(x(:,2)*ones(1,p)-ones(m,1)*knts(:,2)')*diag(1./sigma2));
end
if der==2               % 2nd deriv wrt (x1,x1), (x1,x2) and (x2,x2)
    B2(:,:,1) = exp(-.5*nrms2*diag(1./sigma2)).* ...
        (-(x(:,1)*ones(1,p)-ones(m,1)*knts(:,1)')*diag(1./sigma2)).^2 ...
        + exp(-.5*nrms2*diag(1./sigma2))*(-diag(1./sigma2));
    B2(:,:,2) = exp(-.5*nrms2*diag(1./sigma2)).* ...
        (-(x(:,1)*ones(1,p)-ones(m,1)*knts(:,1)')*diag(1./sigma2)).* ...
        (-(x(:,2)*ones(1,p)-ones(m,1)*knts(:,2)')*diag(1./sigma2));
    B2(:,:,3) = exp(-.5*nrms2*diag(1./sigma2)).* ...
        (-(x(:,2)*ones(1,p)-ones(m,1)*knts(:,2)')*diag(1./sigma2)).^2 ...
        + exp(-.5*nrms2*diag(1./sigma2))*(-diag(1./sigma2));
end
% Normalized basis
sumB0 = sum(B0,2);
NB0 = B0./repmat(sumB0,1,p);
if der==1 || der==2
    sumB1 = sum(B1,2);
    NB1(:,:,1) = (B1(:,:,1)-NB0.*repmat(sumB1(:,:,1),1,p))./repmat(sumB0,1,p);
    NB1(:,:,2) = (B1(:,:,2)-NB0.*repmat(sumB1(:,:,2),1,p))./repmat(sumB0,1,p);
end
if der==2
    sumB2 = sum(B2,2);
    NB2(:,:,1) = (B2(:,:,1)-NB0.*repmat(sumB2(:,:,1),1,p)- ...
        2*NB1(:,:,1).*repmat(sumB1(:,:,1),1,p))./repmat(sumB0,1,p);
    NB2(:,:,2) = (B2(:,:,2)-NB0.*repmat(sumB2(:,:,2),1,p)- ...
        NB1(:,:,1).*repmat(sumB1(:,:,2),1,p)- ...
        NB1(:,:,2).*repmat(sumB1(:,:,1),1,p))./repmat(sumB0,1,p);
    NB2(:,:,3) = (B2(:,:,3)-NB0.*repmat(sumB2(:,:,3),1,p)- ...
        2*NB1(:,:,2).*repmat(sumB1(:,:,2),1,p))./repmat(sumB0,1,p);
end
switch der
    case 0
        B = NB0;
    case 1
        B = NB1;
    case 2
        B = NB2;
    otherwise
        disp('Error -- input DER must be 0, 1 or 2')
        B = [];
end
