function [logf_new,effects_new] = pred_stpp(x_new,y_new,basis_x,basis_y,param)
%[logf_new,effects_new] = pred_stpp(x_new,y_new,basis_x,basis_y,param)
%
% Random-effect prediction for Doubly-Stochastic Spatio-Temporal models
%
% INPUT:
%   x_new: Time points (n_new x 1 cell).
%      Each x_new{i} is (m(i) x 1), the time data for replication i.
%   y_new: Spatial points (n_new x 1 cell).
%      Each y_new{i} is (m(i) x 2), the spatial data for replication i.
%      Each row of y_new{i} are spatial coordinates (longitude/latitude).
%   basis_x: As in STPP.
%   basis_y: As in STPP.
%   param: Output of STPP.
%
% OUTPUT:
%   logf_new: Individual log-densities (n_new x 1).
%   effects_new: Random-effect predictors. Struct with the following fields:
%       z: Individual log-rates (n_new x 1).
%       u: Individual temporal component scores (n_new x p1).
%       v: Individual spatial component scores (n_new x p2).
%
% External calls: BSPL, NGBASIS
%
% Version: March 2019

%% Input check
mssg = in_check(x_new,y_new);
if ~isempty(mssg)
    logf_new = [];
    effects_new = [];
    disp(mssg)
    return
end

%% Data filtering (elimination of data outside RNG)
    % Filtering X's (and associated Y's)
a = basis_x.rng(1);
b = basis_x.rng(2);
n = length(x_new);
for i = 1:n
    jout = x_new{i}<a | x_new{i}>b;
    x_new{i}(jout) = [];
    y_new{i}(jout,:) = [];
end
    % Filtering Y's (and associated X's)
m = zeros(n,1);
for i = 1:n
    jout = ~basis_y.Rdom(y_new{i});
    y_new{i}(jout,:) = [];
    x_new{i}(jout) = [];
    m(i) = length(x_new{i});
end

%% Initialization
    % X basis quantities
ng1 = 300;
q1 = basis_x.or + basis_x.nk;
t = linspace(a,b,ng1);
dt = t(2)-t(1);
knt_x = linspace(a,b,basis_x.nk+2);
B0_x = bspl(t,basis_x.or,knt_x,0);
sumB_x = zeros(n,q1);
for i = 1:n
    B_x_i = bspl(x_new{i},basis_x.or,knt_x,0);
    sumB_x(i,:) = sum(B_x_i,1);
end
    % Y basis quantities
ng2 = 100;
s1 = linspace(basis_y.I1(1),basis_y.I1(2),ng2)';
s2 = linspace(basis_y.I2(1),basis_y.I2(2),ng2)';
ds1 = s1(2)-s1(1);
ds2 = s2(2)-s2(1);
ds = ds1*ds2;
s = [kron(s1,ones(ng2,1)) kron(ones(ng2,1),s2)];
s = s(basis_y.Rdom(s),:);
B0_y = NGbasis(s,basis_y,0);
q2 = size(B0_y,2);
sumB_y = zeros(n,q2);
for i = 1:n
    B_y_i = NGbasis(y_new{i},basis_y,0);
    sumB_y(i,:) = sum(B_y_i,1);
end

%% Main computations
p1 = size(param.C,2);
p2 = size(param.D,2);
if p1==0 && p2==0
    w = zeros(n,1);
    [w,w2,logf_new] = compeff0(sumB_x,sumB_y,param.tau,param.c0,param.d0,...
        param.s2z,dt,ds,B0_x,B0_y,m,w);
    z = w;
    u = [];
    v = [];
else
    S = [param.s2z, param.Szu, param.Szv; ...
        param.Szu', diag(param.s2u), param.Suv; ...
        param.Szv', param.Suv', diag(param.s2v)];
    w = zeros(n,1+p1+p2);
    [w,w2,logf_new] = compeff(sumB_x,sumB_y,param.tau,param.c0,param.d0,...
        param.C,param.D,S,dt,ds,B0_x,B0_y,m,w);
    z = w(:,1);
    u = w(:,2:1+p1);
    v = w(:,1+p1+1:1+p1+p2);
end

%% Output
effects_new = struct('z',z,'u',u,'v',v);

end


%% %%%%%------ AUXILIARY FUNCTIONS

function mssg = in_check(x,y)
mssg = [];
if ~iscell(x)
    mssg = 'Error: X must be cell array';
else
    [mx,nx] = size(x);
    if (mx>1 && nx>1)
        mssg = 'Error: X must be a one-dimensional cell array';
    end
end
if ~iscell(y)
    mssg = 'Error: Y must be cell array';
else
    [my,ny] = size(y);
    if (my>1 && ny>1)
        mssg = 'Error: Y must be a one-dimensional cell array';
    end
end
if length(y)~=length(x)
    mssg = 'X and Y must be cells of the same length N';
end
for i = 1:length(x)
    if length(x{i})~=size(y{i},1)
        mssg = 'x{i} and y{i} must have the same number of rows';
    end
end
end

function [w,w2,logf,flag] = compeff0(sumB_x,sumB_y,tau,c0,d0,s2z,...
    dt,ds,B0_x,B0_y,m,w_ini)
% Computes random effects and log-pdf's using Laplace approx 
%  for mean-only model
n = size(sumB_x,1);
logf = zeros(n,1);
w = w_ini;
w2 = zeros(n,1);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i);
    D_gi = 0;
    H_gi = 1;
    gi = -Inf;
    for steps = 1:NR_steps
        if abs(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            lmbi_x = exp(B0_x*c0);
            lmbi_y = exp(B0_y*d0);
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            r = exp(tau+wL);
            D_gi = -r*I_x*I_y + m(i) - wL/s2z;
            H_gi = -r*I_x*I_y - 1/s2z;
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*log(r) ...
                + sumB_x(i,:)*c0 + sumB_y(i,:)*d0 ...
                -.5*log(2*pi) - .5*log(s2z) - .5*wL^2/s2z;
        else
            flag = true;
        end
    end
    if abs(H_gi)>eps
        SL = 1/(-H_gi);
    else
        SL = 1/(-last_H);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (1/2)*log(2*pi) + .5*log(SL);
    w(i) = wL;
    w2(i) = SL + wL^2;
end
end

function [w,w2,logf,flag] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
    S,dt,ds,B0_x,B0_y,m,w_ini)
%Computes random effects and log-pdf's using Laplace approximations
n = size(sumB_x,1);
p1 = size(C,2);
p2 = size(D,2);
p = p1+p2+1;
Phi_x = B0_x*C;
Phi_y = B0_y*D;
logf = zeros(n,1);
w = w_ini;
w2 = zeros(p,p,n);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    gi = -Inf;
    for steps = 1:NR_steps
        if rcond(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            zL = wL(1);
            uL = wL(2:p1+1);
            vL = wL(p1+2:p);
            r = exp(tau+zL);
            lmbi_x = exp(B0_x*(c0+C*uL'));
            lmbi_y = exp(B0_y*(d0+D*vL'));
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            I_x_der = (lmbi_x'*Phi_x)*dt;
            I_y_der = (lmbi_y'*Phi_y)*ds;
            I_x_hess = Phi_x'*((lmbi_x*ones(1,p1)).*Phi_x)*dt;
            I_y_hess = Phi_y'*((lmbi_y*ones(1,p2)).*Phi_y)*ds;
            D_gi = [-r*I_x*I_y + m(i),...
                -r*I_y*I_x_der + sumB_x(i,:)*C, ...
                -r*I_x*I_y_der + sumB_y(i,:)*D] ...
                - wL/S;
            H_gi = [-r*I_x*I_y, -r*I_y*I_x_der, -r*I_x*I_y_der; ...
                -r*I_y*I_x_der', -r*I_y*I_x_hess, -r*I_x_der'*I_y_der;... 
                -r*I_x*I_y_der', -r*I_y_der'*I_x_der, -r*I_x*I_y_hess]...
                - S\eye(p);
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*(tau+zL) ...
                + sumB_x(i,:)*(c0+C*uL') + sumB_y(i,:)*(d0+D*vL') ...
                -.5*p*log(2*pi) - .5*log(det(S)) - .5*wL*(S\wL');
        else
            flag = true;
        end
    end
    if rcond(H_gi)>eps
        SL = (-H_gi)\eye(p);
    else
        SL = (-last_H)\eye(p);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (p/2)*log(2*pi) + .5*log(det(SL));
    w(i,:) = wL;
    w2(:,:,i) = SL + wL'*wL;
end
end
