function [param,effects,logf] = stpp(x,y,basis_x,basis_y,p1,p2,sm,per,itmax)
%[param,effects,logf] = stpp(x,y,basis_x,basis_y,p1,p2,sm,per,itmax)
%
% Doubly-Stochastic Model for Replicated Spatio-Temporal Point Processes
%
% INPUT:
%   x: Observed time points of the process (n x 1 cell).
%      Each x{i} is (m(i) x 1), the time data from replication i.
%   y: Observed spatial points of the process (n x 1 cell).
%      Each y{i} is (m(i) x 2), the spatial data from replication i.
%      Each row of y{i} are spatial coordinates (longitude/latitude).
%   basis_x: B-spline basis parameters for time process.
%     Struct with the following fields:
%       rng: Time range (1 x 2 vector).
%       or: Spline order (integer; 4 is cubic splines).
%       nk: Number of knots (integer). Knots will be equally spaced.
%   basis_y: Parameters for spatial normalized Gaussian kernel basis.
%             Struct with the following fields:
%       I1: Range for first coordinate of y's (2 x 1 vector).
%       I2: Range for second coordinate of y's (2 x 1 vector).
%           (Note: the region of interest R must be contained in I1 x I2,
%            but the precise domain R is defined by Rdom below).
%       Rdom: Domain indicator function (function handle).
%             For an (m x 2) matrix A, rng(A) must return a logical 
%             (m x 1) vector indicating whether each row of A is in the
%             region of interest R.
%       nk: Number of knots on each axis (integer).
%           Equally spaced knots on each axis are used. In principle this
%           gives nk^2 centroids on I1 x I2 for the Gaussian kernels, but 
%           those outside R are eliminated; the q2 that remain are the
%           dimension of the basis. As "sigma" for each Gaussian  
%           kernel we take half the distance between its centroid and the  
%           next closest centroid.
%   p1: Number of components for temporal intensity process (integer>=0).
%   p2: Number of components for spatial intensity process (integer>=0).
%   sm: Smoothing parameters. Struct with the following fields:
%       muX: Parameter for the mean of temporal intensity process.
%       pcX: Parameter for the components of temporal intensity process.
%       muY: Parameter for the mean of spatial intensity process.
%       pcY: Parameter for the components of spatial intensity process.
%       All parameters ar (scalar>=0).
%   per: Periodicity flag for temporal intensity functions (logical).
%        Use per = 1 if intensity functions are periodic.
%   itmax: Maximum number of iterations (integer>=1).
%
% OUTPUT:
%   param: Parameter estimators. Struct with the following fields:
%       tau: Mean of log-rate (scalar).
%       s2z: Variance of log-rate (scalar).
%       c0: Basis coefficients of temporal mean (q1 x 1).
%       C: Basis coefficients of temporal components (q1 x p1).
%       s2u: Variances of temporal components (p1 x 1).
%       d0: Basis coefficients of spatial mean (q2 x 1).
%       D: Basis coefficients of spatial components (q2 x p2).
%           (Note: the basis coefficients d0 and D are arranged 
%            in the same order as the q2 basis functions computed by 
%            NGbasis when called with the parameters in "basis_y" above).
%       s2v: Variances of spatial components (p2 x 1).
%       Szu: Cross-covariance of log-rate and temporal components (1 x p1).
%       Szv: Cross-covariance of log-rate and spatial components (1 x p2).
%       Suv: Cross-covariance of temporal and spatial components (p1 x p2).
%   effects: Random-effect predictors. Struct with the following fields:
%       z: Individual log-rates (n x 1).
%       u: Individual temporal component scores (n x p1).
%       v: Individual spatial component scores (n x p2).
%   logf: Individual log-densities (n x 1).
%
% External calls: BSPL, NGBASIS
%
% Version: February 2019

%% Input check
mssg = in_check(x,y,p1,p2);
if ~isempty(mssg)
    param = [];
    effects = [];
    logf = [];
    disp(mssg)
    return
end

%% Data filtering (elimination of data outside RNG)
    % Filtering X's (and associated Y's)
a = basis_x.rng(1);
b = basis_x.rng(2);
n = length(x);
for i = 1:n
    jout = x{i}<a | x{i}>b;
    x{i}(jout) = [];
    y{i}(jout,:) = [];
end
    % Filtering Y's (and associated X's)
m = zeros(n,1);
for i = 1:n
    jout = ~basis_y.Rdom(y{i});
    y{i}(jout,:) = [];
    x{i}(jout) = [];
    m(i) = length(x{i});
end

%% Initialization
    % X basis quantities
ng1 = 300;
q1 = basis_x.or + basis_x.nk;
t = linspace(a,b,ng1);
dt = t(2)-t(1);
knt_x = linspace(a,b,basis_x.nk+2);
B0_x = bspl(t,basis_x.or,knt_x,0);
J0_x = (B0_x'*B0_x)*dt;
if basis_x.or>2
    B2_x = bspl(t,basis_x.or,knt_x,2);
    J2_x = (B2_x'*B2_x)*dt;
else
    J2_x = zeros(q1,q1);
end
sumB_x = zeros(n,q1);
for i = 1:n
    B_x_i = bspl(x{i},basis_x.or,knt_x,0);
    sumB_x(i,:) = sum(B_x_i,1);
end
Bab0 = bspl([a b],basis_x.or,knt_x,0);
Bab1 = bspl([a b],basis_x.or,knt_x,1);
AP = [Bab0(1,:)-Bab0(2,:); Bab1(1,:)-Bab1(2,:)]';
N_per = null(AP');
P_per = N_per*N_per';
a0_x = (sum(B0_x,1)*dt)';
N_0_x = null(a0_x');
P_0_x = N_0_x*N_0_x'; 
aux_BB_x = (kron(ones(q1,1),B0_x').*kron(B0_x',ones(q1,1)));
    % Y basis quantities
[J0_y,J2_y] = compJ(basis_y);
ng2 = 100;
s1 = linspace(basis_y.I1(1),basis_y.I1(2),ng2)';
s2 = linspace(basis_y.I2(1),basis_y.I2(2),ng2)';
ds1 = s1(2)-s1(1);
ds2 = s2(2)-s2(1);
ds = ds1*ds2;
s = [kron(s1,ones(ng2,1)) kron(ones(ng2,1),s2)];
s = s(basis_y.Rdom(s),:);
B0_y = NGbasis(s,basis_y,0);
q2 = size(B0_y,2);
sumB_y = zeros(n,q2);
for i = 1:n
    B_y_i = NGbasis(y{i},basis_y,0);
    sumB_y(i,:) = sum(B_y_i,1);
end
a0_y = (sum(B0_y,1)*ds)';
N_0_y = null(a0_y');
P_0_y = N_0_y*N_0_y'; 
aux_BB_y = (kron(ones(q2,1),B0_y').*kron(B0_y',ones(q2,1)));

%% Mean-only model
disp(' ')
disp(' -----> Computing mean-only model')
disp(' ')
tau = mean(log(max(m,1)/((b-a)*areaB(basis_y))));
s2z = var(log(max(m,1)/((b-a)*areaB(basis_y))));
c0 = zeros(q1,1);
d0 = zeros(q2,1);
w = zeros(n,1);
[w,w2,logf] = compeff0(sumB_x,sumB_y,tau,c0,d0,s2z,dt,ds,B0_x,B0_y,m,w);
Pen = sm.muX*c0'*J2_x*c0 + sm.muY*d0'*J2_y*d0;
OF = mean(logf) - Pen;
disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
iter = 0;
err = 1;
while err>1e-3 && iter<itmax
    iter = iter + 1;
    s2z_0 = s2z;
    % Parameter updates
    c0 = update_c0(sumB_x,tau,c0,d0,B0_x,dt,B0_y,ds,sm,J2_x,per,AP);
    d0 = update_d0(sumB_y,tau,c0,d0,B0_x,dt,B0_y,ds,sm,J2_y);
    tau = update_tau0(m,w,c0,d0,B0_x,dt,B0_y,ds);
    s2z = mean(w2);
    % Random-effect and OF updates
    [w,w2,logf] = compeff0(sumB_x,sumB_y,tau,c0,d0,s2z,dt,ds,B0_x,B0_y,m,w);
    Pen = sm.muX*c0'*J2_x*c0 + sm.muY*d0'*J2_y*d0;
    OF = mean(logf) - Pen;
    % Convergence check
    err = abs(s2z-s2z_0)/s2z_0;
    disp(['Iteration: ' num2str(iter) ', OF: ' num2str(OF) ...
            ', Error: ' num2str(err)])
end

%% Sequential PC estimation
    % Components up to p = min(p1,p2)
C = [];
s2u = [];
D = [];
s2v = [];
Szu = [];
Szv = [];
Suv = [];
z = w;
u = [];
v = [];
p = min(p1,p2);
for ic = 1:p
    disp(' ')
    disp(['---> Computing component ' num2str(ic)])
    disp(' ')
    % Initial estimators
    if ic==1
        PC = P_0_x;
        PD = P_0_y;
    else
        NC = null(C(:,1:ic-1)'*J0_x);
        PC_ort = NC*NC';
        PC = P_0_x*PC_ort;
        ND = null(D(:,1:ic-1)'*J0_y);
        PD_ort = ND*ND';
        PD = P_0_y*PD_ort;
    end
    if per
        PC = P_per*PC;
    end
    Cnew = PC*ones(q1,1);
    Dnew = PD*ones(q2,1);
    C = [C, Cnew/sqrt(Cnew'*J0_x*Cnew)];
    u = [u, zeros(n,1)];
    D = [D, Dnew/sqrt(Dnew'*J0_y*Dnew)];
    v = [v, zeros(n,1)];
    if ic==1
        s2u = s2z/4;
        s2v = s2z/4;
    else
        s2u(ic) = s2u(ic-1)/4;
        s2v(ic) = s2v(ic-1)/4;
    end
    Szu = [Szu, 0];
    Szv = [Szv, 0];
    Suv = [Suv, zeros(size(Suv,1),1); zeros(1,size(Suv,2)), 0];
    S = [s2z, Szu, Szv; ...
        Szu', diag(s2u), Suv; ...
        Szv', Suv', diag(s2v)];
    w = [z,u,v];
    [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,S,dt,ds,...
        B0_x,B0_y,m,w);
    Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
        + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
    OF = mean(logf) - Pen;
    disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
    % Main iterations
    iter = 0;
    err = 1;
    iu = 1+1:1+ic;
    iv = 1+ic+1:1+ic+ic;
    while (err>1e-3 && iter<itmax) || (iter<=2)
        iter = iter + 1;
        S0 = S;
        % Parameter updates
        C = updateC(sumB_x,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
            J0_x,J2_x,per,AP,aux_BB_x,a0_x);
        D = updateD(sumB_y,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
            J0_y,J2_y,aux_BB_y,a0_y);
        tau = update_tau(m,w,c0,C,d0,D,B0_x,dt,B0_y,ds);
        S = mean(w2,3);
        [U1,L1] = svd(S(iu,iu));
        [U2,L2] = svd(S(iv,iv));
        Rot = [1, zeros(1,ic), zeros(1,ic); ...
            zeros(ic,1), U1', zeros(ic,ic); ...
            zeros(ic,1), zeros(ic,ic), U2'];
        S = Rot*S*Rot';
        s2z = S(1,1);
        s2u = diag(S(iu,iu));
        s2v = diag(S(iv,iv));
        Szu = S(1,iu);
        Szv = S(1,iv);
        Suv = S(iu,iv);
        C = C*U1;
        u = u*U1;
        D = D*U2;
        v = v*U2;
        w = [z,u,v];
        % Random-effect and OF updates
        [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
            S,dt,ds,B0_x,B0_y,m,w);
        Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
            + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
        OF = mean(logf) - Pen;
        % Convergence check
        err = abs(norm(abs(S(:))-abs(S0(:))))/norm(abs(S0(:)));
        disp(['Iteration: ' num2str(iter) ', OF: ' num2str(OF) ...
            ', Error: ' num2str(err)])
    end
    z = w(:,1);
    u = w(:,iu);
    v = w(:,iv);
end

    % Remaining X components if p1>p
if p1>p
    for ic = p+1:p1
        disp(' ')
        disp(['---> Computing component ' num2str(ic) ' of X'])
        disp(' ')
        % Initial estimators
        NC = null(C(:,1:ic-1)'*J0_x);
        PC_ort = NC*NC';
        PC = P_0_x*PC_ort;
        if per
            PC = P_per*PC;
        end
        Cnew = PC*ones(q1,1);
        C = [C, Cnew/sqrt(Cnew'*J0_x*Cnew)];
        u = [u, zeros(n,1)];
        s2u(ic) = s2u(ic-1)/4;
        Szu = [Szu, 0];
        Suv = [Suv; zeros(1,size(Suv,2))];
        S = [s2z, Szu, Szv; ...
            Szu', diag(s2u), Suv; ...
            Szv', Suv', diag(s2v)];
        w = [z,u,v];
        [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,S,dt,ds,...
            B0_x,B0_y,m,w);
        Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
            + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
        OF = mean(logf) - Pen;
        disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
        % Main iterations
        iter = 0;
        err = 1;
        iu = 1+1:1+ic;
        iv = 1+ic+1:1+ic+p2;
        while (err>1e-3 && iter<itmax) || (iter<=2)
            iter = iter + 1;
            S0 = S;
            % Parameter updates
            C = updateC(sumB_x,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
                J0_x,J2_x,per,AP,aux_BB_x,a0_x);
            tau = update_tau(m,w,c0,C,d0,D,B0_x,dt,B0_y,ds);
            S = mean(w2,3);
            [U1,L1] = svd(S(iu,iu));
            [U2,L2] = svd(S(iv,iv));
            Rot = [1, zeros(1,ic), zeros(1,p2); ...
                zeros(ic,1), U1', zeros(ic,p2); ...
                zeros(p2,1), zeros(p2,ic), U2'];
            S = Rot*S*Rot';
            s2z = S(1,1);
            s2u = diag(S(iu,iu));
            s2v = diag(S(iv,iv));
            Szu = S(1,iu);
            Szv = S(1,iv);
            Suv = S(iu,iv);
            C = C*U1;
            u = u*U1;
            D = D*U2;
            v = v*U2;
            w = [z,u,v];
            % Random-effect and OF updates
            [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
                S,dt,ds,B0_x,B0_y,m,w);
            Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
                + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
            OF = mean(logf) - Pen;
            % Convergence check
            err = abs(norm(abs(S(:))-abs(S0(:))))/norm(abs(S0(:)));
            disp(['Iteration: ' num2str(iter) ', OF: ' num2str(OF) ...
                ', Error: ' num2str(err)])
        end
        z = w(:,1);
        u = w(:,iu);
        v = w(:,iv);
    end
end

    % Remaining Y components if p2>p
if p2>p
    for ic = p+1:p2
        disp(' ')
        disp(['---> Computing component ' num2str(ic) ' of Y'])
        disp(' ')
        % Initial estimators
        ND = null(D(:,1:ic-1)'*J0_y);
        PD_ort = ND*ND';
        PD = P_0_y*PD_ort;
        Dnew = PD*ones(q2,1);
        D = [D, Dnew/sqrt(Dnew'*J0_y*Dnew)];
        v = [v, zeros(n,1)];
        s2v(ic) = s2v(ic-1)/4;
        Szv = [Szv, 0];
        Suv = [Suv, zeros(size(Suv,1),1)];
        S = [s2z, Szu, Szv; ...
            Szu', diag(s2u), Suv; ...
            Szv', Suv', diag(s2v)];
        w = [z,u,v];
        [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,S,dt,ds,...
            B0_x,B0_y,m,w);
        Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
            + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
        OF = mean(logf) - Pen;
        disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
        % Main iterations
        iter = 0;
        err = 1;
        iu = 1+1:1+p1;
        iv = 1+p1+1:1+p1+ic;
        while (err>1e-3 && iter<itmax) || (iter<=2)
            iter = iter + 1;
            S0 = S;
            % Parameter updates
            D = updateD(sumB_y,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
                J0_y,J2_y,aux_BB_y,a0_y);
            tau = update_tau(m,w,c0,C,d0,D,B0_x,dt,B0_y,ds);
            S = mean(w2,3);
            [U1,L1] = svd(S(iu,iu));
            [U2,L2] = svd(S(iv,iv));
            Rot = [1, zeros(1,p1), zeros(1,ic); ...
                zeros(p1,1), U1', zeros(p1,ic); ...
                zeros(ic,1), zeros(ic,p1), U2'];
            S = Rot*S*Rot';
            s2z = S(1,1);
            s2u = diag(S(iu,iu));
            s2v = diag(S(iv,iv));
            Szu = S(1,iu);
            Szv = S(1,iv);
            Suv = S(iu,iv);
            C = C*U1;
            u = u*U1;
            D = D*U2;
            v = v*U2;
            w = [z,u,v];
            % Random-effect and OF updates
            [w,w2,logf] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
                S,dt,ds,B0_x,B0_y,m,w);
            Pen = sm.muX*c0'*J2_x*c0 + sm.pcX*trace(C'*J2_x*C)...
                + sm.muY*d0'*J2_y*d0 + sm.pcY*trace(D'*J2_y*D);
            OF = mean(logf) - Pen;
            % Convergence check
            err = abs(norm(abs(S(:))-abs(S0(:))))/norm(abs(S0(:)));
            disp(['Iteration: ' num2str(iter) ', OF: ' num2str(OF) ...
                ', Error: ' num2str(err)])
        end
        z = w(:,1);
        u = w(:,iu);
        v = w(:,iv);
    end
end

%% Output
param = struct('tau',tau,'s2z',s2z,'c0',c0,'C',C,'s2u',s2u, ...
    'd0',d0,'D',D,'s2v',s2v,'Szu',Szu,'Szv',Szv,'Suv',Suv);
if p1==0 && p2==0
    effects = struct('z',w,'u',[],'v',[]);
else
    effects = struct('z',w(:,1),'u',w(:,2:p1+1),'v',w(:,p1+2:p1+p2+1));
end

end


%% %%%%%------ AUXILIARY FUNCTIONS

function mssg = in_check(x,y,p1,p2)
mssg = [];
if min(p1,p2)==0 && (p1>0 || p2>0)
    mssg = 'This program only works for both p''s positive or both zero';
end
if ~iscell(x)
    mssg = 'Error: X must be cell array';
else
    [mx,nx] = size(x);
    if (mx>1 && nx>1)
        mssg = 'Error: X must be a one-dimensional cell array';
    end
end
if ~iscell(y)
    mssg = 'Error: Y must be cell array';
else
    [my,ny] = size(y);
    if (my>1 && ny>1)
        mssg = 'Error: Y must be a one-dimensional cell array';
    end
end
if length(y)~=length(x)
    mssg = 'X and Y must be cells of the same length N';
end
for i = 1:length(x)
    if length(x{i})~=size(y{i},1)
        mssg = 'x{i} and y{i} must have the same number of rows';
    end
end
end

function a = areaB(basis_y)
% Computes area of Rdom
ng2 = 100;
s1 = linspace(basis_y.I1(1),basis_y.I1(2),ng2)';
s2 = linspace(basis_y.I2(1),basis_y.I2(2),ng2)';
ds1 = s1(2)-s1(1);
ds2 = s2(2)-s2(1);
ds = ds1*ds2;
s = [kron(s1,ones(ng2,1)) kron(ones(ng2,1),s2)];
s = s(basis_y.Rdom(s),:);
a = size(s,1)*ds;
end

function [w,w2,logf,flag] = compeff0(sumB_x,sumB_y,tau,c0,d0,s2z,...
    dt,ds,B0_x,B0_y,m,w_ini)
% Computes random effects and log-pdf's using Laplace approx 
%  for mean-only model
n = size(sumB_x,1);
logf = zeros(n,1);
w = w_ini;
w2 = zeros(n,1);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i);
    D_gi = 0;
    H_gi = 1;
    gi = -Inf;
    for steps = 1:NR_steps
        if abs(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            lmbi_x = exp(B0_x*c0);
            lmbi_y = exp(B0_y*d0);
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            r = exp(tau+wL);
            D_gi = -r*I_x*I_y + m(i) - wL/s2z;
            H_gi = -r*I_x*I_y - 1/s2z;
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*log(r) ...
                + sumB_x(i,:)*c0 + sumB_y(i,:)*d0 ...
                -.5*log(2*pi) - .5*log(s2z) - .5*wL^2/s2z;
        else
            flag = true;
        end
    end
    if abs(H_gi)>eps
        SL = 1/(-H_gi);
    else
        SL = 1/(-last_H);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (1/2)*log(2*pi) + .5*log(SL);
    w(i) = wL;
    w2(i) = SL + wL^2;
end
end

function c0 = update_c0(sumB_x,tau,c0,d0,B0_x,dt,B0_y,ds,sm,J2_x,per,AP)
% Updates c0 for mean-only model
[n,q] = size(sumB_x);
a0 = (sum(B0_x,1)*dt)';
lmb_x = exp(B0_x*c0);
lmb_y = exp(B0_y*d0);
I_y = sum(lmb_y)*ds;
mu = B0_x*c0;
IlmbBB = (B0_x'*((lmb_x*ones(1,q)).*B0_x))*dt;
Q = exp(tau)*I_y*IlmbBB + 2*sm.muX*J2_x;
b = -exp(tau)*I_y*(B0_x'*(lmb_x.*(1-mu)))*dt + mean(sumB_x,1)';
if per
    sol_sys = [Q, [AP a0]; [AP a0]', zeros(3)]\[b;0;0;0];
else
    sol_sys = [Q, a0; a0', 0]\[b;0];
end
c0 = sol_sys(1:q);
end

function d0 = update_d0(sumB_y,tau,c0,d0,B0_x,dt,B0_y,ds,sm,J2_y)
% Updates d0 for mean-only model
[n,q] = size(sumB_y);
a0 = (sum(B0_y,1)*ds)';
lmb_x = exp(B0_x*c0);
I_x = sum(lmb_x)*dt;
lmb_y = exp(B0_y*d0);
mu = B0_y*d0;
IlmbBB = (B0_y'*((lmb_y*ones(1,q)).*B0_y))*ds;
Q = exp(tau)*I_x*IlmbBB + 2*sm.muY*J2_y;
b = -exp(tau)*I_x*(B0_y'*(lmb_y.*(1-mu)))*ds + mean(sumB_y,1)';
sol_sys = [Q, a0; a0', 0]\[b;0];
d0 = sol_sys(1:q);
end

function tau = update_tau0(m,w,c0,d0,B0_x,dt,B0_y,ds)
% Updates tau for mean-only model
lmb_x = exp(B0_x*c0);
lmb_y = exp(B0_y*d0);
I_x = sum(lmb_x)*dt;
I_y = sum(lmb_y)*ds;
tau = log(mean(m)/mean(exp(w)*I_x*I_y));
end

function [w,w2,logf,flag] = compeff(sumB_x,sumB_y,tau,c0,d0,C,D,...
    S,dt,ds,B0_x,B0_y,m,w_ini)
%Computes random effects and log-pdf's using Laplace approximations
n = size(sumB_x,1);
p1 = size(C,2);
p2 = size(D,2);
p = p1+p2+1;
Phi_x = B0_x*C;
Phi_y = B0_y*D;
logf = zeros(n,1);
w = w_ini;
w2 = zeros(p,p,n);
flag = false;
NR_steps = 5;
for i = 1:n
    wL = w(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    gi = -Inf;
    for steps = 1:NR_steps
        if rcond(H_gi)>eps
            last_H = H_gi;
            last_gi = gi;
            last_wL = wL;
            wL = wL - D_gi/H_gi;
            zL = wL(1);
            uL = wL(2:p1+1);
            vL = wL(p1+2:p);
            r = exp(tau+zL);
            lmbi_x = exp(B0_x*(c0+C*uL'));
            lmbi_y = exp(B0_y*(d0+D*vL'));
            I_x = sum(lmbi_x)*dt;
            I_y = sum(lmbi_y)*ds;
            I_x_der = (lmbi_x'*Phi_x)*dt;
            I_y_der = (lmbi_y'*Phi_y)*ds;
            I_x_hess = Phi_x'*((lmbi_x*ones(1,p1)).*Phi_x)*dt;
            I_y_hess = Phi_y'*((lmbi_y*ones(1,p2)).*Phi_y)*ds;
            D_gi = [-r*I_x*I_y + m(i),...
                -r*I_y*I_x_der + sumB_x(i,:)*C, ...
                -r*I_x*I_y_der + sumB_y(i,:)*D] ...
                - wL/S;
            H_gi = [-r*I_x*I_y, -r*I_y*I_x_der, -r*I_x*I_y_der; ...
                -r*I_y*I_x_der', -r*I_y*I_x_hess, -r*I_x_der'*I_y_der;... 
                -r*I_x*I_y_der', -r*I_y_der'*I_x_der, -r*I_x*I_y_hess]...
                - S\eye(p);
            gi = -r*I_x*I_y - gammaln(m(i)+1) + m(i)*(tau+zL) ...
                + sumB_x(i,:)*(c0+C*uL') + sumB_y(i,:)*(d0+D*vL') ...
                -.5*p*log(2*pi) - .5*log(det(S)) - .5*wL*(S\wL');
        else
            flag = true;
        end
    end
    if rcond(H_gi)>eps
        SL = (-H_gi)\eye(p);
    else
        SL = (-last_H)\eye(p);
        gi = last_gi;
        wL = last_wL;
    end
    logf(i) = gi + (p/2)*log(2*pi) + .5*log(det(SL));
    w(i,:) = wL;
    w2(:,:,i) = SL + wL'*wL;
end
end

function C = updateC(sumB_x,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
    J0_x,J2_x,per,AP,aux_BB_x,a0_x)
% Updates component coefficients C(:,end)
[n,q] = size(sumB_x);
p1 = size(C,2);
p2 = size(D,2);
z = w(:,1);
u = w(:,2:p1+1);
v = w(:,p1+2:p1+1+p2);
r = exp(tau+z);
u2 = squeeze(w2(1+p1,1+p1,:));
lmb_x = exp(B0_x*(c0*ones(1,n)+C*u'));
lmb_y = exp(B0_y*(d0*ones(1,n)+D*v'));
I_y = sum(lmb_y,1)*ds;
uPhi = (B0_x*C(:,p1))*u(:,p1)';
b = -dt*B0_x'*((lmb_x.*(1-uPhi))*(r.*I_y'.*u(:,p1)))/n ...
    + (sumB_x'*u(:,p1))/n;
%aux = (kron(ones(q,1),B0_x').*kron(B0_x',ones(q,1)));
vM2 = dt*aux_BB_x*(lmb_x*(r.*I_y'.*u2))/n;
M2 = reshape(vM2,[q q]);
Q = M2 + 2*sm.pcX*J2_x;
if p1>1
    if per
        A = [(C(:,1:p1-1)'*J0_x)' AP a0_x];
        f = zeros(size(A,2),1);
    else
        A = [(C(:,1:p1-1)'*J0_x)' a0_x];
        f = zeros(size(A,2),1);
    end
else
    if per
        A = [AP a0_x];
        f = zeros(size(A,2),1);
    else
        A = a0_x;
        f = 0;
    end
end
sol_sys = [Q A; A' zeros(size(A,2))]\[b;f];
C_new = sol_sys(1:q);
C(:,p1) = C_new/sqrt(C_new'*J0_x*C_new);
end

function D = updateD(sumB_y,w,w2,tau,c0,C,d0,D,B0_x,dt,B0_y,ds,sm,...
    J0_y,J2_y,aux_BB_y,a0_y)
% Updates component coefficients D(:,end)
[n,q] = size(sumB_y);
p1 = size(C,2);
p2 = size(D,2);
z = w(:,1);
u = w(:,2:p1+1);
v = w(:,p1+2:p1+1+p2);
r = exp(tau+z);
v2 = squeeze(w2(p1+1+p2,p1+1+p2,:));
lmb_x = exp(B0_x*(c0*ones(1,n)+C*u'));
I_x = sum(lmb_x,1)*dt;
lmb_y = exp(B0_y*(d0*ones(1,n)+D*v'));
vPhi = (B0_y*D(:,p2))*v(:,p2)';
b = -ds*B0_y'*((lmb_y.*(1-vPhi))*(r.*I_x'.*v(:,p2)))/n ...
    + (sumB_y'*v(:,p2))/n;
%aux = (kron(ones(q,1),B0_y').*kron(B0_y',ones(q,1)));
vM2 = ds*aux_BB_y*(lmb_y*(r.*I_x'.*v2))/n;
M2 = reshape(vM2,[q q]);
Q = M2 + 2*sm.pcY*J2_y;
if p2>1
    A = [(D(:,1:p2-1)'*J0_y)' a0_y];
    f = zeros(size(A,2),1);
else
    A = a0_y;
    f = 0;
end
sol_sys = [Q A; A' zeros(size(A,2))]\[b;f];
D_new = sol_sys(1:q);
D(:,p2) = D_new/sqrt(D_new'*J0_y*D_new);
end

function tau = update_tau(m,w,c0,C,d0,D,B0_x,dt,B0_y,ds)
% Updates rate parameter tau
n = length(m);
p1 = size(C,2);
p2 = size(D,2);
z = w(:,1);
u = w(:,2:p1+1);
v = w(:,p1+2:p1+1+p2);
lmb_x = exp(B0_x*(c0*ones(1,n)+C*u'));
I_x = sum(lmb_x,1)*dt;
lmb_y = exp(B0_y*(d0*ones(1,n)+D*v'));
I_y = sum(lmb_y,1)*ds;
tau = log(sum(m)/sum(exp(z)'.*I_x.*I_y));
end

function [J0,J2] = compJ(basis)
% Computes matrices J0 and J2 for roughness penalty terms
% J0 is the Gram matrix of the basis functions
% Internal Jik are the Gram matrices of the 2nd deriv wrt (ti,tk)
% J2 = J11 + 2*J12 + J22
m = 100;
g1 = linspace(basis.I1(1),basis.I1(2),m)';
g2 = linspace(basis.I2(1),basis.I2(2),m)';
d1 = g1(2)-g1(1);
d2 = g2(2)-g2(1);
xg_0 = [kron(g1,ones(m,1)) kron(ones(m,1),g2)];
xg_0 = xg_0(basis.Rdom(xg_0),:);
B0_0 = NGbasis(xg_0,basis,0);
q = size(B0_0,2);
J0 = zeros(q,q);
J11 = zeros(q,q);
J12 = zeros(q,q);
J22 = zeros(q,q);
for j = 1:m
    xg = [g1(j)*ones(m,1) g2];
    xg = xg(basis.Rdom(xg),:);
    B0 = NGbasis(xg,basis,0);
    J0 = J0 + d1*d2*(B0'*B0);
    B2 = NGbasis(xg,basis,2);
    J11 = J11 + d1*d2*(B2(:,:,1)'*B2(:,:,1));
    J12 = J12 + d1*d2*(B2(:,:,2)'*B2(:,:,2));
    J22 = J22 + d1*d2*(B2(:,:,3)'*B2(:,:,3));
end
J2 = J11 + 2*J12 + J22;
end
